// -*-c++-*-
// This file is part of the webmpd front-end for MPD music player
//
// Copyright (C) 2006 Hannes Reich (hannes@skynet.ie)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
//

////////////////////////////////////////////////////////////////
// Functions for manipulating HTML element nodes

function setElementDisplay(element, isDisplayed)
{
    element.style.display = isDisplayed?"":"none";
}

function removeElementClass(element, classNameToRemove)
{
    element.className = element.className.replace( 
        regExpForClass(classNameToRemove),
	' '
	);
}

function addElementClass(element, classNameToAdd)
{
    if(!elementHasClass(element, classNameToAdd))
      element.className = [element.className, classNameToAdd].join(' ');
}

function elementHasClass(element, classNameToTest)
{
    return element.className.match(regExpForClass(classNameToTest));
}

function regExpForClass(classNameToTest)
{
    if(!regExpForClass.regExpCache[classNameToTest])
        regExpForClass.regExpCache[classNameToTest] = new RegExp('(^| +)' + classNameToTest + '($| +)');
    return regExpForClass.regExpCache[classNameToTest];
}
regExpForClass.regExpCache = new Object;

function getElementById(id)
{
    var e = document.getElementById(id);
    if(!e)
        log.warn("Failed to find element with id '" + id + "'");
    return e;
}

function findElementWithClass(rootElement, className)
{
    if(rootElement.className && elementHasClass(rootElement, className))
	return rootElement;
    for(var child = rootElement.firstChild; child; child = child.nextSibling)
    {
        if(child.className || child.hasChildNodes()) // avoid text nodes
	{
	    var foundElement = findElementWithClass(child, className);
	    if(foundElement)
		return foundElement;
	}
    }
    return null;
}

function findElementsByClass(rootElement, classNameList)
{
    var children = new Object;
    var classNames = classNameList.split(/\s+/);
    for(var i=0; i<classNames.length; i++)
    {
	children[classNames[i]] = findElementWithClass(rootElement, classNames[i]);
	if(!children[classNames[i]])
	    log.warning("Failed to find element with class '"+className+"'");
    }
    return children;
}

function getElementTextChild(element)
{
    for(var child = element.firstChild; child; child = child.nextSibling)
      if(child.nodeType == 3) // 3 == Node.TEXT_NODE (hardcoded for IE)
        return child;
    log.info("No text child node found, inserting new one");
    var textNode = element.ownerDocument.createTextNode("");
    if(element.hasChildNodes())
	return element.insertBefore(textNode, element.firstChild);

    return element.appendChild(textNode);
}

function setElementTextContent(element, text)
{
    getElementTextChild(element).nodeValue = text;
}

function getElementPosition(element, dimension)
{
        /* Algorithm from http://www.quirksmode.org/js/findpos.html */
	var currentPosition = 0;
	if (element.offsetParent)
	{
		while (element.offsetParent)
		{
		    currentPosition += element[dimension == "x" ? "offsetLeft" : "offsetTop"];
		    element = element.offsetParent;
		}
	}
	else if (element[dimension])
		currentPosition += element[dimension];
	return currentPosition;
}

function getElementXPosition (element) {
    return getElementPosition(element, "x");
}
function getElementYPosition (element) {
    return getElementPosition(element, "y");
}
