#!/usr/bin/perl -wT
# This file is part of the webmpd front-end for MPD music player
#
# Copyright (C) 2006 Hannes Reich (hannes@skynet.ie)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

use strict;
use sigtrap;

use CGI;
use Encode ('decode', 'encode');

use lib "."; # Change this to the full path of the dir containing this
             # script if MPD.pm is not found (depends on your http server)
use MPD;
use JSON;

# MPD connection details
use constant MPD_HOST => "localhost";
use constant MPD_PORT => "6600";

sub in_list($@);
sub get_arguments($$);

# explicit host and port below prevents taint warning that would
# result from allowing them to depend on environment variables.
my $mpd = new MPD(MPD_HOST, MPD_PORT);
my $cgi = new CGI;

print $cgi->header("text/json; encoding=".$cgi->charset());

my $command = $cgi->param('command');
my %output = ("command" => $command);

if(in_list($command,
	   qw(add playid search set_repeat clear deleteid
	   load lsinfo next pause play playlist_changes prev rm save seek set_fade
	   set_random set_volume shuffle stop)
	  ))
  {
    my @arguments = get_arguments($cgi, $command);
    my @value = $mpd->$command(@arguments);
    $output{"arguments"} = \@arguments;

    if("playlist_changes" eq $command)
      {
        $output{"value"} = {@value};
      }
    else
      {
        $output{"value"} = \@value;
      }

    my @error = $mpd->get_error();
    if($error[0])
      {
	$output{"error"} = { "message" => $error[1], "command" => $error[2]?$error[2]:$command };
      }
  }
else
  {
    $output{"error"} = { "message" => "Unknown command", "command" => $command };
  }

#print objToJson(\%output, {"pretty" => 1});
print encode($cgi->charset, objToJson(\%output));

exit;

sub in_list($@)
  {
    my($target, @list) = @_;
    for(@list)
      {
	return 1
	  if $_ eq $target;
      }
    return 0;
  }

sub get_arguments($$)
  {
    my($cgi, $command) = @_;
    my @arguments;
    while(defined (my $p = $cgi->param($#arguments + 1)))
      {
        $p = decode($cgi->charset(), $p);
        if (in_list($command, "playid", "deleteid", "set_volume", "seek", "playlist_changes")) {
          $p =~ s/[^\d]//g;
        }
        push @arguments, $p;
      }
    # TODO validate argument counts and types
    return @arguments;
  }
