function Scheduler()
{
    this.nextId = 0;
    this.queue = [];
    this.schedule_bound = bindFunction(this.schedule, this);
}

Scheduler.timesliceLength = 200;
Scheduler.idleLength = 100;

Scheduler.prototype = {
    createTask : function (task, state) {
        this.queue.push({execute: task, state: state, id: this.nextId});
        // log.info("task " + this.nextId + " added. queue length: " + this.queue.length);
        return this.nextId++;
    },
    killTask : function (idToKill) {
        var oldqueue = this.queue;
        this.queue = [];
        for(var t in oldqueue)
            if(oldqueue[t].id != idToKill)
                this.queue.push(oldqueue[t]);
    },
    // functions for use in the task code
    isTimeSliceExpired: function () {
        return (new Date()).getTime() >= this.timesliceEnd;
    },
    requestRestart : function () {
        this.queue.push(this.current);
    },
    saveState : function (state) {
        this.current.state = state;
    },
    getSavedState : function () {
        return this.current.state;
    },

    schedule : function () {
        // log.info("tick. queue length: " + this.queue.length);
        this.current = this.queue.shift();
        if(this.current)
        {
            this.timesliceEnd = (new Date()).getTime() + Scheduler.timesliceLength;
            this.current.execute();
        }
        this.timerId = window.setTimeout(this.schedule_bound, Scheduler.idleLength);
    },

    stop : function () {
        window.clearTimeout(this.timerId);
    },

    start : function () {
        this.schedule();
    }
};
