// -*-ecmascript-*-
// This file is part of the webmpd front-end for MPD music player
//
// Copyright (C) 2006 Hannes Reich (hannes@skynet.ie)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
//

function setLocalLink(domElement, linkTarget)
{
    // following a local link in konqueror causes the state iframe to
    // stop loading, so don't set the href (TODO find a better solution)
    if(navigator.appName.match("Konqueror"))
        domElement.href = "javascript:void(0)";
    else
        domElement.href = linkTarget;

    // This avoids the lag of waiting for the location poll timer to fire
    domElement.onclick = bindFunction(
        locationWatcher.setLocation, locationWatcher, [linkTarget]
        );
}

function formatTime(seconds)
{
  if(isNaN(seconds))
    return "";
  var s = seconds%60;
  if(s<10) 
    s = '0' + s;
  return Math.floor(seconds/60) + ':' + s;
}

function basename(path)
{
 // The perl JSON code in the backend produces numeric values for
 // filenames consisting only of numbers, so we need the toString()
 // here.
 return path.toString().replace(/^.*\//, "");
}

function setToggleElementClass(e, value)
{
    removeElementClass(e, 'toggle' + (0!=value?'Off':'On'));
    addElementClass(e, 'toggle' + (0!=value?'On':'Off'));
}

function forcePageReflow()
{
    var bml = document.body.style.marginLeft;
    document.body.style.marginLeft = "1px";
    document.body.style.marginLeft = bml;
}

function scrollToElement(element) 
{
    var offset = stickable.getBodyOffset();

    var visibleTop = window.scrollY + offset;
    if(!window.scrollY)
        visibleTop = 0; // workaround for konqueror < 3.5.1
    var visibleHeight = window.innerHeight - offset;
    var visibleBottom = visibleTop + visibleHeight;

    var elementTop = getElementYPosition(element);
    var elementHeight = element.offsetHeight;
    var elementBottom = elementTop + elementHeight;

    var targetPosition;
    if( elementHeight > visibleHeight
        || elementTop < visibleTop
        || elementBottom > visibleBottom
        )
        window.scrollTo(window.scrollX, elementTop - offset);
}

////////////////////////////////////////////////////////////////
function Observable(initalValue)
{
    this.value = initalValue;
    this.observers = new Object();
    this.sequenceNumber = 0;
}

Observable.prototype.setValue = function (newValue)
{
    if(newValue !== this.value)
    {
        this.value = newValue;
        for(var n in this.observers)
        {
            this.observers[n].callback(newValue);
            if(this.observers[n].isOneShot)
                delete this.observers[n];
        }
    }
}

Observable.prototype.getValue = function () {
    return this.value;
}

Observable.prototype.addObserver = function (callback, isOneShot)
{
    var s = this.sequenceNumber++;
    this.observers[s] = {"callback": callback, "isOneShot": isOneShot};
    if(undefined !== this.value && !isOneShot)
        callback(this.value)
    return s;
}

Observable.prototype.addOneShotObserver = function (callback)
{
    return this.addObserver(callback, true);
}

Observable.prototype.removeObserver = function (id)
{
    delete this.observers[id];
}

Observable.prototype.hasObservers = function () {
    for(var n in this.observers)
        return true;
    return false;
}


////////////////////////////////////////////////////////////////
function ElementWrapper(element)
{
    if(!element)
        log.error("No element passed to ElementWrapper");
    this.domElement = element;
    this.domSubElements = new Object();
}

ElementWrapper.prototype.getSubElement = function(className)
{
    if(!this.domSubElements[className])
        this.domSubElements[className]
           = findElementWithClass(this.domElement, className);
    return this.domSubElements[className];
}

////////////////////////////////////////////////////////////////
function TemplateableElementWrapper(templateId, element)
{
    if(!element)
    {
	var templateNode = getElementById(templateId);
	element = templateNode.cloneNode(true);
	element.id = ""; // try to avoid duplicate IDs
    }
    getSuper(arguments).apply(this, [element]);
}

inheritFrom(TemplateableElementWrapper, ElementWrapper);

////////////////////////////////////////////////////////////////
// Display of player status

function PlayerView(domElement) {
    getSuper(arguments).apply(this, arguments);

    this.data = {artist : "", title : "", file : "", state : ""};

    this.volumeProgressBar = new ProgressBar(this.getSubElement("volumeProgressBar"),
					     bindFunction(backendController.set_volume,
							  backendController
						 )
	                                    );
    this.trackProgressBar = new ProgressBar(this.getSubElement("trackProgressBar"),
					    bindFunction(backendController.seek,
							 backendController
						)
					);

    this.currentSongPosition = new PlaylistPosition("#", this.getSubElement("songIndex"));
    this.playlistLength = new PlaylistPosition("", this.getSubElement("playlistLength"));
}

PlayerView.prototype.observeModel = function (model)
{
    var functionMap = {
        "song": this.setSongIndex,
        "playlistlength": this.setPlaylistLength,
        "song.Title": this.setTitle,
        "song.file": this.setFilename,
        "song.Artist": this.setArtist,
        "song.Album": this.setAlbum,
        "trackTimes": this.setTrackTimes,
        "state": this.setState,
        "volume": this.setVolume,
        "xfade": this.setXfade,
        "random": this.setRandomState,
        "repeat": this.setRepeatState
    };

    for(var a in functionMap)
    {
        model.getAttributeObservable(a)
            .addObserver(
                bindFunction(functionMap[a], this)
        );
    }
}

PlayerView.prototype.setArtist = function (t) {
    this.data.artist = t;
    setElementTextContent(this.getSubElement("artist"), t);
    this.updateDocumentTitle();
}
PlayerView.prototype.setTitle = function (t) {
    this.data.title = t;
    setElementTextContent(this.getSubElement("trackName"), this.getTrackName());
    this.updateDocumentTitle();
}
PlayerView.prototype.setFilename = function (t) {
    this.data.file = t;
    setElementTextContent(this.getSubElement("trackName"), this.getTrackName());
    this.updateDocumentTitle();
}
PlayerView.prototype.getTrackName = function () {
    if("" == this.data.title)
      return this.data.file;
    else
      return this.data.title;
}
PlayerView.prototype.setAlbum = function (t) {
    setElementTextContent(this.getSubElement("album"), t);
}
PlayerView.prototype.setSongIndex = function (p) {
    this.currentSongPosition.setValue("" === p ? 0 : (toNumber(p) + 1));
}
PlayerView.prototype.setPlaylistLength = function (l) {
    this.playlistLength.setValue(l);
}

PlayerView.prototype.setRepeatState = function (state) {
    setToggleElementClass(this.getSubElement("repeatState"), state);
}
PlayerView.prototype.setRandomState = function (state) {
    setToggleElementClass(this.getSubElement("randomState"), state);
}
PlayerView.prototype.setXfade = function (value) {
    setToggleElementClass(this.getSubElement("xfadeState"), value);
}

PlayerView.prototype.setState = function (t) {
    this.data.state = t;
    setElementTextContent(this.getSubElement("stateText"), t);
    removeElementClass(this.domElement, 'playerPlaying');
    removeElementClass(this.domElement, 'playerPaused');
    removeElementClass(this.domElement, 'playerStopped');
    addElementClass(this.domElement, 'player' + t);
    this.updateDocumentTitle();
}

PlayerView.prototype.setVolume = function (value) {
    this.volumeProgressBar.setValue(value,100);
    this.volumeProgressBar.setText(value);
}

PlayerView.prototype.setTrackTimes = function (times)
{
    var elapsed = times[0];
    var total = times[1];
    if(0 == total) /* happens with streams */
      total = elapsed;

    this.trackProgressBar.setValue(elapsed, total);
    this.trackProgressBar.setLeftText(formatTime(elapsed));
    this.trackProgressBar.setRightText(formatTime(total-elapsed));
}

PlayerView.prototype.updateDocumentTitle = function() {
    var t = "MPD (" + this.data.state + "): ";
    if(this.data.artist)
      t += this.data.artist + " - ";
    if(this.data.title)
      t += this.data.title;
    else
      if(this.data.file)
        t += this.data.file;
    this.domElement.ownerDocument.title = t;
}

inheritFrom(PlayerView, ElementWrapper);

////////////////////////////////////////////////////////////////
function ObservableModel()
{
    this.attributeData = {};
}

ObservableModel.prototype.getAttribute = function (key) {
    return this.getAttributeObservable(key).getValue();
}
ObservableModel.prototype.setAttribute = function (key, value) {
    return this.getAttributeObservable(key).setValue(value);
}
ObservableModel.prototype.getAttributeObservable = function (key) {
    if(!this.attributeData[key])
        this.attributeData[key] = new Observable();
    return this.attributeData[key];
}
ObservableModel.prototype.copyModel = function (m) {
    var a;
    for(a in this.attributeData)
        if(!(a in m.attributeData))
            this.setAttribute(a, undefined);
    for(a in m.attributeData)
        this.setAttribute(a, m.attributeData[a].getValue());
}

////////////////////////////////////////////////////////////////
function PlayerStatusModel()
{
    getSuper(arguments).apply(this, arguments);
}

PlayerStatusModel.prototype.observeBackendMonitor = function(monitor)
{
    monitor.getObservable().addObserver(
        bindFunction(this.acceptMonitorUpdate, this)
        );
}

PlayerStatusModel.prototype.acceptMonitorUpdate = function(keyValuePair)
{
    var key = keyValuePair[0];
    var value = keyValuePair[1];
    if("time" === key)
    {
        var times = value.split(':');
        if(2!=times.length)
            times = [0,0];
        key = "trackTimes";
        value = times;
    }
    if("state" === key)
      value =
        "play" == value ? "Playing" :
        "pause" == value ? "Paused" :
	"Stopped";

    this.setAttribute(key, value);
}

inheritFrom(PlayerStatusModel, ObservableModel)

////////////////////////////////////////////////////////////////

function ListModel()
{
    this.length = new Observable(0);
    this.dataComplete = new Observable(false);
}

ListModel.prototype.getItemModel = null;

ListModel.prototype.createItem = null;

////////////////////////////////////////////////////////////////

function SparseListModel()
{
    getSuper(arguments).apply(this, arguments);
    this.itemModels = {};
}

SparseListModel.prototype.getItemModel = function (index) {
    index = toNumber(index);
    if(!this.itemModels[index])
        this.itemModels[index] = this.createItem();
    if(index >= this.length.getValue())
        this.length.setValue(index+1);
    return this.itemModels[index];
}

SparseListModel.prototype.setLength = function (length) {
    length = toNumber(length);
    if(length < this.length.getValue())
        for(var i in this.itemModels)
            if(i >= length)
                delete this.itemModels[i];
    this.length.setValue(length);
}

SparseListModel.prototype.getLength = function () {
    return this.length.getValue();
}

SparseListModel.prototype.forEachItem = function (callback) {
    for(var index in this.itemModels)
      callback(this.itemModels[index]);
}

inheritFrom(SparseListModel, ListModel);

////////////////////////////////////////////////////////////////

function SelectableSparseListModel()
{
    getSuper(arguments).apply(this, arguments);
}

SelectableSparseListModel.prototype.selectAll = function () {
    this.forEachItem(function (i) { i.setAttribute("isSelected", true); });
}
SelectableSparseListModel.prototype.invertSelection = function () {
    this.forEachItem(function (i) { 
        i.setAttribute("isSelected", !i.getAttribute("isSelected"));
    });
}
SelectableSparseListModel.prototype.forEachSelectedItem = function (callback) {
    this.forEachItem(function (i) { 
        if(i.getAttribute("isSelected"))
            callback(i);
    });
}

inheritFrom(SelectableSparseListModel, SparseListModel);

////////////////////////////////////////////////////////////////

function PlaylistModel()
{
    getSuper(arguments).apply(this, arguments);
    this.pendingRequests = new Object;
    this.latestRequestedVersion = 0;
    this.backend = null;
    this.currentSongPosition = 0;
}

PlaylistModel.prototype.observePlayerStatusModel = function (model) {
    model.getAttributeObservable("playlist").addObserver(
         bindFunction(this.updateToNewVersion, this)
         );
    model.getAttributeObservable("playlistlength").addObserver(
         bindFunction(this.setLength, this)
        );
    model.getAttributeObservable("song").addObserver(
        bindFunction(this.setCurrentSongPosition, this)
        );
}

PlaylistModel.prototype.useBackend = function (backend) {
    if(this.backend)
      log.warn("calling useBackend twice is unlikely to work, as it doesn't invalidate old data");
    this.backend = backend;
}

PlaylistModel.prototype.createItem = function () {
    var m = new PlaylistItemModel();
    return m;
}

PlaylistModel.prototype.setCurrentSongPosition = function (index) {
    index = toNumber(index);
    if(index >= this.getLength())
      this.setLength(index+1);
    this.getItemModel(this.currentSongPosition).setAttribute("isCurrentItem", false);
    this.currentSongPosition = index;
    this.getItemModel(this.currentSongPosition).setAttribute("isCurrentItem", true);
}

PlaylistModel.prototype.updateToNewVersion = function (newVersion) {
    log.info("new playlist version seen: " + newVersion);
    this.dataComplete.setValue(false);

    if(newVersion > this.latestRequestedVersion)
    {
        var seq = this.backend.playlist_changes(
            this.latestRequestedVersion,
            bindFunction(this.acceptListChangeData,
                         this, [newVersion]
                )
            );
      this.pendingRequests[seq] = "";
      this.latestRequestedVersion = newVersion;
    }
}

PlaylistModel.prototype.acceptListChangeData = function (version, data) {
    this.backend.defaultCommandCallback(data);
    delete this.pendingRequests[data.sequenceNumber];
    log.info("got playlist change data for version " + version);
    scheduler.createTask(
        bindFunction(this.processListChangeDataTask, this),
        {version: version, data: data});
}

PlaylistModel.prototype.processListChangeDataTask = function () {
    var version = scheduler.getSavedState().version;
    var data = scheduler.getSavedState().data;

    for(var index in data.value)
    {
        // TODO: can't check index against list length here, as the
        // relative ordering of length and version number updates from
        // MPD is uncertain. By not checking, we risk accepting
        // obsolete updates for items past the end of the list.
        // Possible fix: don't let getItemModel update the length
        // value -- then we can store data past the end, but noone
        // will look at it.

        this.getItemModel(index).acceptUpdate(version, data.value[index]);
        delete data.value[index];

	if(scheduler.isTimeSliceExpired())
	{
	    log.info("PlaylistModel.acceptListChangeData() exit, deferring further processing");
            scheduler.requestRestart();
	    return;
	}
    }

    this.dataComplete.setValue(isEmpty(this.pendingRequests));

    log.info("PlaylistModel.acceptListChangeData() exit, processing complete.");
    return;
}

inheritFrom(PlaylistModel, SelectableSparseListModel);

////////////////////////////////////////////////////////////////
function PlaylistItemModel() {
    getSuper(arguments).apply(this, arguments);
}

PlaylistItemModel.prototype.acceptUpdate = function (version, update) {
    if(this.getAttribute("version") > version)
    {
        log.info("dropping obsolete change data for playlist item");
        return;
    }

    var defaultValues =  {
        "isCurrentItem": false,
        "isBeingDeleted": false,
        "isSelected": false,
        "Time": 0
    };
    
    var stickyAttributes =  {
        "isCurrentItem": "" // current item status is managed independently of updates
    };


    var self = this;
    function hasStickyValue(key) {
        return key in stickyAttributes 
            && key in self.attributeData
            && undefined !== self.attributeData[key].value;
    }

    var key;
    // overwrite old data
    for(key in this.attributeData)
    {
        if(hasStickyValue(key)) continue;
        if(key in update) continue;
        if(key in defaultValues) continue;
        this.setAttribute(key, undefined);
    }

    // set default values
    for(key in defaultValues)
    {
        if(hasStickyValue(key)) continue;
        if(key in update) continue;
        this.setAttribute(key, defaultValues[key]);
    }

    // set new values
    for(key in update)
      this.setAttribute(key, update[key]);

    this.setAttribute("version", version);

    return;
}

inheritFrom(PlaylistItemModel, ObservableModel);

////////////////////////////////////////////////////////////////
function PlaylistPositionIndex() {
    this.lastSeenPlaylistLength = 0;
    this.lastSeenFilesByPosition = {};
    this.fileObservables = {};
    this.filesRequiringScan = {};
    this.scanRequestPending = false;
}

PlaylistPositionIndex.prototype.observePlaylistModel = function (model) {
    this.playlistModel = model;
    model.length.addObserver(bindFunction(this.setPlaylistLength, this));
}

PlaylistPositionIndex.prototype.setPlaylistLength = function (newLength) {
    newLength = toNumber(newLength);
    var i;

    // if the list has grown, add observers to the new file attributes
    for(i=this.lastSeenPlaylistLength; i<newLength; i++)
         this.playlistModel.getItemModel(i).getAttributeObservable("file")
             .addObserver(
                 bindFunction(this.setFileAtPosition, this, [i])
                 );

    // if the list shrunk, take account of the disappeared items
    for(i=newLength; i<this.lastSeenPlaylistLength; i++)
        if(this.lastSeenFilesByPosition[i])
            this.considerUpdatingFile(this.lastSeenFilesByPosition[i], 
                                      function (p) {delete p[i]} );
    

    this.lastSeenPlaylistLength = newLength;
}

PlaylistPositionIndex.prototype.setFileAtPosition = function (position, newFile) {
    if(this.lastSeenFilesByPosition[position])
        this.considerUpdatingFile(this.lastSeenFilesByPosition[position], 
                                  function (p) {delete p[position]});

    this.considerUpdatingFile(newFile, 
                              function (p) {p[position] = ""});

    this.lastSeenFilesByPosition[position] = newFile;
}

PlaylistPositionIndex.prototype.considerUpdatingFile = function (file, updateFunction) {
    // (expensively) modify a file's observable, but only if
    // there's someone watching it
    if(file && this.fileObservables[file])
    {
        var observable = this.fileObservables[file];
        if(observable.hasObservers())
        {
            var p = shallowClone(observable.getValue());
            updateFunction(p);
            observable.setValue(p);
        }
        else 
            delete this.fileObservables[file];
    }
}

// wait this long before starting a playlist scan, in case more files
// get added to this.filesRequiringScan in the meantime
PlaylistPositionIndex.playlistScanDelay = 500;

PlaylistPositionIndex.prototype.getFileObservable = function (file) {
    if(!this.fileObservables[file])
    {
        this.fileObservables[file] = new Observable({});
        this.filesRequiringScan[file] = "";
        if(!this.scanRequestPending)
        {
            window.setTimeout(bindFunction(this.startPlaylistScan, this),
                              PlaylistPositionIndex.playlistScanDelay);
            this.scanRequestPending = true;
        }
    }
    return this.fileObservables[file];
}

PlaylistPositionIndex.prototype.startPlaylistScan = function () {
    this.scanRequestPending = false;
    var files = this.filesRequiringScan;
    this.filesRequiringScan = {};
    scheduler.createTask(bindFunction(this.playlistScanTask, this),
                         {files: files, offset: 0});
}
PlaylistPositionIndex.prototype.playlistScanTask = function () {
    var files = scheduler.getSavedState().files;
    var offset = scheduler.getSavedState().offset;
    for(var i=offset; i<this.playlistModel.getLength(); i++)
    {
        for(var file in files)
            if(file === this.playlistModel.getItemModel(i).getAttribute("file"))
                this.considerUpdatingFile(file, function (p) {p[i] = ""});
        if(scheduler.isTimeSliceExpired())
        {
            scheduler.saveState({files: files, offset: i+1});
            scheduler.requestRestart();
            return;
        }
    }
}


////////////////////////////////////////////////////////////////
// Generic dynamic list
function ContiguousList() {
    this.items = new Array();
    this.expectedCallbackSequenceNumber = -1;
}

ContiguousList.prototype.setLength = function (length, fastShorten) {
    if(fastShorten)
        // warning: only use this option if skipping popItem will do no harm
        this.items.splice(length, this.getLength() - length);
    else
        while(this.getLength() > length)
	    this.popItem();

    while(this.getLength() < length)
	this.pushItem(this.createItem());
}

ContiguousList.prototype.getLength = function () {
    return this.items.length;
}

ContiguousList.prototype.pushItem = function (item) {
    this.items.push(item);
}

ContiguousList.prototype.popItem = function () {
    return this.items.pop();
}

ContiguousList.prototype.getItem = function (index) {
    return this.items[index];
}

// factory method used by setLength
ContiguousList.prototype.createItem = function () {
    return {};
}

ContiguousList.prototype.forEachItem = function (callback) {
    for(var index=0; index<this.items.length; index++)
      callback(this.items[index]);
}

ContiguousList.prototype.setExpectedCallbackSequenceNumber = function (n) {
    this.expectedCallbackSequenceNumber = n;
}

ContiguousList.PROCESSING_ABORTED = 0;
ContiguousList.PROCESSING_COMPLETE = 1;
ContiguousList.PROCESSING_DEFERRED = 2;

ContiguousList.prototype.acceptCallbackData = function (data)
{
    if(data.error)
    {
	log.error("MPD backend error: " +data.error.message);
	return ContiguousList.PROCESSING_ABORTED;
    }

    scheduler.createTask(bindFunction(this.processCallbackDataTask, this), {data:data, offset:0});
    return ContiguousList.PROCESSING_DEFERRED;
}

ContiguousList.prototype.processCallbackDataTask = function ()
{
    var data = scheduler.getSavedState().data;

    if(data.sequenceNumber != this.expectedCallbackSequenceNumber)
    {
      log.info("Detected reply to obsolete data request. Discarding.");
      return ContiguousList.PROCESSING_ABORTED;
    }

    var offset = scheduler.getSavedState().offset;
    if(!offset)
	this.setLength(0);

    log.info("ContiguousList.processCallbackDataTask() entry at offset " + offset);

    for(var i=offset; i<data.value.length; i++)
    {
	var entry = data.value[i];
        var d = this.createItemFromCallbackData(entry);
        if(d)
           this.pushItem(d);

	if(scheduler.isTimeSliceExpired())
	{
	    log.info("ContiguousList.processCallbackDataTask() exit, deferring further processing at offset " + i);
            scheduler.saveState({data:data, offset: i+1});
            scheduler.requestRestart();
	    return ContiguousList.PROCESSING_DEFERRED;
	}
    }

    log.info("ContiguousList.processCallbackDataTask() exit, processing complete.");
    return ContiguousList.PROCESSING_COMPLETE;
}

ContiguousList.prototype.createItemFromCallbackData = function(entry) {
    return null;
}

////////////////////////////////////////////////////////////////

function ListView(domElement) {
    if(!domElement)
	domElement = document.createElement("ol");
    ElementWrapper.apply(this, arguments);
    ContiguousList.apply(this, arguments);
    this.domListElement = this.domElement;
}

ListView.prototype.pushItem = function (item) {
    getSuper(arguments).apply(this, arguments);
    this.domListElement.appendChild(item.domElement);
}

ListView.prototype.popItem = function () {
    var i = getSuper(arguments).apply(this, arguments);
    if(i && i.domElement.parentNode)
        this.domListElement.removeChild(i.domElement);
    return i;
}

ListView.prototype.createItem = function () {
    return new ElementWrapper(this.domElement.ownerDocument.createElement("li"));
}

ListView.prototype.processCallbackDataTask = function ()
{
    var status = getSuper(arguments).apply(this, arguments);
    if(ContiguousList.PROCESSING_COMPLETE === status)
        removeElementClass(this.domElement, "updating"); // set by subclasses - TODO tidy this up

    return status;
}

ListView.prototype.createItemFromCallbackData = function(entry) {
    if(entry.file)
    {
        var d = new FileView();
        for(var a in entry)
            d.setAttribute(a, entry[a]);
        return d;
    }

    return getSuper(arguments).apply(this, arguments);
}

inheritFrom(ListView, ElementWrapper);
inheritFrom(ListView, ContiguousList);

////////////////////////////////////////////////////////////////
function ScrollableListModelView(domElement) {
    getSuper(arguments).apply(this, arguments);
    this.model = undefined;
    this.modelOffset = 0;
    this.pageOverlap = 1;

    this.getSubElement("pageLengthSelect").onchange = 
        bindFunction(this.pageLengthChanged, this);
    this.pageLengthChanged();

    this.getSubElement("pageUpLink").onclick = bindFunction(this.pageUp, this);
    this.getSubElement("pageDownLink").onclick = bindFunction(this.pageDown, this);
}

ScrollableListModelView.prototype.observeModel = function (model)
{
    this.model = model;
    model.length.addObserver(bindFunction(this.acceptModelLength, this));
    model.dataComplete.addObserver(bindFunction(this.setIsDataComplete, this));
    this.updateHiddenItemCounts();
}

ScrollableListModelView.prototype.acceptModelLength = function (l) {
    this.adjustModelOffset(this.modelOffset)
}

ScrollableListModelView.prototype.setIsDataComplete = function (complete) {
    (complete?removeElementClass:addElementClass)(this.domElement, "updating");
}

ScrollableListModelView.prototype.popItem = function () {
    var poppedItem = getSuper(arguments).apply(this);
    if(poppedItem)
        poppedItem.unobserve();
    return poppedItem;
}

ScrollableListModelView.prototype.pushItem = function (item) {
    item.observeModel(
        this.model.getItemModel(this.modelIndexForItem(
                                    this.getLength()
                                    )
            )
        );
    getSuper(arguments).apply(this, arguments);
}

ScrollableListModelView.prototype.modelIndexForItem = function (index) {
    return index + this.modelOffset;
}

ScrollableListModelView.prototype.adjustModelOffset = function (offset) {
    if(this.pageLength)
    {
        // reduce offset if the model is too short
        if(offset + this.pageLength > this.model.getLength())
            offset = max(0, this.model.getLength() - this.pageLength);
        this.modelOffset = offset;

        // reduce view length if necessary
        this.setLength(max(0,
                           min(
                               this.pageLength, 
                               this.model.getLength() - offset
                               )
                           )
            );
    }
    else
    {
        this.modelOffset = 0;
        this.setLength(this.model.getLength());
    }

    for(var i=0; i<this.items.length; i++)
        this.items[i].observeModel(this.model.getItemModel(this.modelIndexForItem(i)));

    this.updateHiddenItemCounts();
}

ScrollableListModelView.prototype.pageLengthChanged = function () {
    var v = this.getSubElement("pageLengthSelect").value;
    if("all" === v)
        this.pageLength = 0;
    else
        this.pageLength = toNumber(v);

    if(this.model)
        this.adjustModelOffset(this.modelOffset);
}

ScrollableListModelView.prototype.pageDown = function () {
    this.adjustModelOffset(this.modelOffset + this.pageLength - this.pageOverlap);
}

ScrollableListModelView.prototype.pageUp = function () {
    this.adjustModelOffset(max(0, this.modelOffset - (this.pageLength - this.pageOverlap)));
}

ScrollableListModelView.prototype.updateHiddenItemCounts = function () {
    var above = max(0, this.modelOffset);
    var below = max(0, this.model.getLength() - this.modelOffset - this.pageLength);
    if(!this.pageLength)
        below = 0;
    setElementDisplay(this.getSubElement("pageUpLink"), above != 0);
    setElementTextContent(this.getSubElement("hiddenAboveCount"), above);
    setElementDisplay(this.getSubElement("pageDownLink"), below != 0);
    setElementTextContent(this.getSubElement("hiddenBelowCount"), below);
}

inheritFrom(ScrollableListModelView, ListView);

////////////////////////////////////////////////////////////////
function SelectableListView(domElement) {
    getSuper(arguments).apply(this, arguments);
    this.getSubElement("selectAllLink").onclick = bindFunction(this.selectAll, this);
    this.getSubElement("invertSelectionLink").onclick = bindFunction(this.invertSelection, this);
}

SelectableListView.prototype.selectAll = function () {
    this.model.selectAll();
}
SelectableListView.prototype.invertSelection = function () {
    this.model.invertSelection();
}

inheritFrom(SelectableListView, ScrollableListModelView);

////////////////////////////////////////////////////////////////
// Playlist 

function PlaylistView(domElement) {
    getSuper(arguments).apply(this, arguments);
    this.domListElement = this.getSubElement("songList");

    this.getSubElement("shuffleLink").onclick = bindFunction(backendController.shuffle,
                                                             backendController,
                                                             [null]);
    this.getSubElement("clearLink").onclick = bindFunction(this.clear, this, [null]);
    this.getSubElement("deleteLink").onclick = bindFunction(this.deleteSelected, this);

    this.getSubElement("addLink").onclick = function () {
        var url = window.prompt("Enter URL to add");
        if(url) backendController.add(url);
    };
}

PlaylistView.prototype.createItem = function () {
  return new PlaylistItemView();
}

PlaylistView.prototype.showItem = function (position) {
    if(!this.model.dataComplete.getValue())
    {
        log.info("Trying to scroll playlist, but data not yet complete, deferring.");
        this.model.dataComplete.addOneShotObserver(bindFunction(this.showItem,
                                                                this, [position]
                                                       ));
        return;
    }

    log.info("scrolling to playlist item " + position);

    if(0 == position)
    {
        this.adjustModelOffset(0);
        scrollToElement(this.domListElement);
    }
    else
    {
        if(position <= this.modelOffset || position > this.modelOffset + this.pageLength)
            this.adjustModelOffset(position - 1);
    
        var viewIndex = position - this.modelOffset - 1;
    
        if(viewIndex < this.items.length)
            scrollToElement(this.items[viewIndex].domElement);
        else
            log.warn("Trying to scroll to playlist item " + position
                     + " but playlist is only " + this.modelIndexForItem(this.items.length)
                     + " items long");
    }
}

PlaylistView.prototype.deleteSelected =  function () {
    var fileIds = new Array();
    this.model.forEachSelectedItem(function(i) {
	fileIds.push(i.getAttribute("Id"));
        i.setAttribute("isBeingDeleted", true);
    });
    backendController.deleteid(fileIds);
}

PlaylistView.prototype.clear =  function () {
    this.forEachItem(function(i) {
        i.setIsBeingDeleted(true);
    });
    backendController.clear();
}

inheritFrom(PlaylistView, SelectableListView);

////////////////////////////////////////////////////////////////
function TrackView(templateId)
{
    getSuper(arguments).apply(this, arguments);
    this.data = new Object;
    this.observerIds = {};
}

TrackView.prototype.setText = function (text)
{
    setElementTextContent(this.getSubElement("textContainer"), text);
}

TrackView.prototype.getFilename = function ()
{
    if(this.data.file)
      return this.data.file;
    return null;
}

TrackView.prototype.setAttribute = function (key, value)
{
    this.data[key] = value;
    switch(key)
    {
	case "file":
	    if(value && 0 == value.indexOf("http://"))
	    {
		var parentDir = value
		    .replace(/\/[^\/]*$/, "/")   // strip last directory
		    .replace(/\:\d*\/?$/, "/"); // strip port number on root
		this.getSubElement("browseLink").onclick = null;
		this.getSubElement("browseLink").href = parentDir;
	    }
	    else
	    {
                setLocalLink(this.getSubElement("browseLink"),
                             "#browse;" + escape(value)
                    );
	    }
	case "Artist":
	case "Album":
	case "Title":
	case "Name":
	    if(this.data.Name)// some streams use this
		this.setText(this.data.Name);
            else
                if(this.data.Title)
                    this.setText(this.data.Artist
                                 + " - " 
                                 + this.data.Title);
                else
                    if(this.data.file)
                        this.setText(this.data.file);
	    break;
	case "Time":
	    setElementTextContent(this.getSubElement("trackTime"), formatTime(value)); 
	    break;
	case "Id":
	    break;
	default:
	    log.info("Unexpected attribute in TrackView.setAttribute: '"+key+"'->'"+value+"'");
    }
}

TrackView.prototype.getAttribute = function (key)
{
    return this.data[key];
}

TrackView.prototype.observeModel = function (model) {
    if(model === this.model)
        return;
    this.unobserve();
    for(var a in {"file": "",
                  "Artist": "",
                  "Album": "",
                  "Title": "",
                  "Name": "",
                  "Time": "",
                  "Id": ""})
    {
        this.setAttribute(a, "");
        this.observerIds[a] =  model.getAttributeObservable(a).addObserver(
            bindFunction(this.setAttribute, this, [a])
            );
    }
    this.model = model;
}

TrackView.prototype.unobserve = function () {
    if(this.model)
    {
        for(var a in this.observerIds)
            this.model.getAttributeObservable(a).removeObserver(this.observerIds[a]);
    }
}

inheritFrom(TrackView, TemplateableElementWrapper);

////////////////////////////////////////////////////////////////
function SelectableTrackView(templateId)
{
    getSuper(arguments).apply(this, arguments);

    this.getSubElement("selectInput").onchange 
        = bindFunction(this.checkboxChanged,
                       this);
}

SelectableTrackView.prototype.observeModel = function (model) {
    if(model !== this.model)
    {
        getSuper(arguments).apply(this, arguments);
        this.observerIds.isSelected = model.getAttributeObservable("isSelected")
            .addObserver(
                bindFunction(this.setIsSelected, this)
                );
    }
}

SelectableTrackView.prototype.setIsSelected = function (x) {
    this.getSubElement("selectInput").checked = x;
    (x?addElementClass:removeElementClass)(this.domElement, "selected");
    return x;
}

SelectableTrackView.prototype.checkboxChanged = function () {
    this.model.setAttribute("isSelected",
                            this.getSubElement("selectInput").checked
        );
}

inheritFrom(SelectableTrackView, TrackView);

////////////////////////////////////////////////////////////////
// Single item in the playlist

function PlaylistItemView()
{
    getSuper(arguments).apply(this, ["template_playlistItem"]);
}

PlaylistItemView.prototype.observeModel = function (model) {
    if(model !== this.model)
    {
        getSuper(arguments).apply(this, arguments)
        this.observerIds.isCurrentItem = model.getAttributeObservable("isCurrentItem")
            .addObserver(
                bindFunction(this.setIsCurrentItem, this)
            );
        this.observerIds.isBeingDeleted = model.getAttributeObservable("isBeingDeleted")
            .addObserver(
                bindFunction(this.setIsBeingDeleted, this)
            );
    }
}

PlaylistItemView.prototype.setIsCurrentItem = function (status)
{
    (status?addElementClass:removeElementClass)(this.domElement, "currentSong");
}


PlaylistItemView.prototype.setIsBeingDeleted = function (status)
{
    (status?addElementClass:removeElementClass)(this.domElement, "deleting");
}

PlaylistItemView.prototype.setAttribute = function (key, value)
{
    getSuper(arguments).apply(this, arguments);
    if("Id" == key)
      this.setId(value);
}

PlaylistItemView.prototype.setId = function (id)
{
    // TODO remove this stuff
    this.getSubElement("textContainer").onclick = bindFunction(
	backendController.playid,
	backendController,
	[id, null]
	);
    this.getSubElement("deleteLink").onclick = 
      bindFunction(this.requestDelete, this);
}

PlaylistItemView.prototype.requestDelete = function () {
    if(undefined === this.data.Id)
      log.warn("trying to delete playlist item with no id attribute");
    this.model.setAttribute("isBeingDeleted", true);
    backendController.deleteid(this.data.Id);
}

inheritFrom(PlaylistItemView, SelectableTrackView);

////////////////////////////////////////////////////////////////
// Slider control for track progress and volume

function ProgressBar(domElement, clickCallback) {
    getSuper(arguments).apply(this, [domElement]);
    this.value = 0;
    this.maxValue = 100;
    this.domElement.onclick = bindFunction(this.processClickEvent, this);
    this.clickCallback = clickCallback;
}

// TODO setText() positioning should be updated for  setValue()

ProgressBar.prototype.setValue = function (v, max) {
    this.value = v;
    this.maxValue = max;
    var percent = 100 * this.value / this.maxValue;
    if(isNaN(percent)) percent = 0;
    this.getSubElement("progressBarCompleted").style.width = percent + "%";
}

ProgressBar.prototype.setLeftText = function (t) {
    setElementTextContent(this.getSubElement("progressBarLeftText"), t);
}

ProgressBar.prototype.setRightText = function (t) {
    setElementTextContent(this.getSubElement("progressBarRightText"), t);
}

ProgressBar.prototype.setText = function (t) {
    if(this.value < this.maxValue / 2)
    {
	this.setLeftText('');
	this.setRightText(t);
    }
    else
    {
	this.setLeftText(t);
	this.setRightText('');
    }
}

ProgressBar.prototype.processClickEvent = function(event) {
    if(this.clickCallback)
    {
      var value = this.maxValue
                  * (event.clientX - getElementXPosition(this.domElement))
                  / this.domElement.clientWidth;
      this.clickCallback(value);
    }
}

inheritFrom(ProgressBar, ElementWrapper);

////////////////////////////////////////////////////////////////
// Database directory browser

function DirectoryView(path, domElement) {
    getSuper(arguments).apply(this, ["template_directory", domElement]);

    this.contentView = new DirectoryContentView(this.getSubElement("directoryContent"));

    this.path = path;
    this.expanded = false;
    this.getSubElement("expandLink").onclick = bindFunction(this.toggleExpanded, this);
    this.getSubElement("addLink").onclick = bindFunction(backendController.add, 
						       backendController, 
						       [path, null]);

    this.setText(basename(path));
}

DirectoryView.prototype.getPath = function () {
    return this.path;
}

DirectoryView.prototype.setText = function (text) {
    setElementTextContent(this.getSubElement("textContainer"), text);    
}

DirectoryView.prototype.expand = function () {
 if(!this.expanded)
   this.toggleExpanded();
}

DirectoryView.prototype.toggleExpanded = function () {
    this.expanded = !this.expanded;
    setElementDisplay(this.getSubElement("directoryContent"), this.expanded);
    if(this.expanded)
    {
	this.getSubElement("expandLink").title = "Close directory";
	if(!this.contentView.dataRequested)
	    this.contentView.requestData(this.path);
    }
    else
    {
	this.getSubElement("expandLink").title = "Open directory";
    }
}

DirectoryView.prototype.showFile = function (path) {
    this.expand();
    if(this.path == path)
      return;
    this.contentView.addDataCompleteSubscriber(
	    bindFunction(this.recurseShowFile, this, [path])
	    );
}

DirectoryView.prototype.recurseShowFile = function (path) {
    log.info("Expanding '"+path+"'");

    if(0 != path.indexOf(this.path))
      log.error("Something wrong in DirectoryView.recurseShowFile: '" + path + "' should start with '" + this.path + "'");

    var subpath = path.slice(this.path.length).match(/^\/?[^\/]*/);
    if(!subpath)
      log.error("Something wrong in DirectoryView.recurseShowFile: can't figure out next step to get from '" + this.path + "' to '" + path + "'");

    var nextDirectory = this.path + subpath;
    log.info("Next step from '"+this.path+"' is '"+nextDirectory+"'");

    var subItem = this.contentView.getItemByPath(nextDirectory);
    if(!subItem)
      log.error("Something wrong in DirectoryView.recurseShowFile: can't find item '" + nextDirectory + "'");
 
    scrollToElement(subItem.domElement);

    if(subItem.showFile)
      subItem.showFile(path);
}

inheritFrom(DirectoryView, TemplateableElementWrapper);

////////////////////////////////////////////////////////////////
function FileView() {
    getSuper(arguments).apply(this, ["template_file"]);
    this.playlistPositions = 
	new PlaylistPositionListView(this.getSubElement("playlistPositions"));
}

FileView.prototype.setAttribute = function (key, value) {
    if("file" == key)
      this.setPath(value);
    getSuper(arguments).apply(this, arguments);
}

FileView.prototype.setPath = function (path) {
    this.getSubElement("addLink").onclick 
    = this.getSubElement("textContainer").onclick 
    = bindFunction(backendController.add, backendController, [path, null]);

    this.playlistPositions.setPath(path);
}

inheritFrom(FileView, TrackView);

////////////////////////////////////////////////////////////////
function SavedListsView(domElement) {
    getSuper(arguments).apply(this, arguments);
    this.domListElement = this.getSubElement("songList");
    this.getSubElement("refreshLink").onclick = bindFunction(this.requestData, this);
    this.getSubElement("saveButton").onclick 
        = this.getSubElement("nameInput").onenter
        = bindFunction(this.saveCurrentList, this);
}

SavedListsView.prototype.createItemFromCallbackData = function(entry) {
    if(entry.playlist)
    return new SavedPlaylistView(entry.playlist,
                                   bindFunction(this.requestData, this)
          );
    return null;
}

SavedListsView.prototype.requestData = function ()
{
    addElementClass(this.domElement, "updating");
    this.setExpectedCallbackSequenceNumber(
        backendController.lsinfo("", bindFunction(this.acceptCallbackData, this))
        );
}

SavedListsView.prototype.isCallbackDataItemInteresting = function(item)
{
    return item.playlist ? true : false;
}

SavedListsView.prototype.saveCurrentList = function() {
    var name = this.getSubElement("nameInput").value;
    for(var i=0; i<this.items.length; i++)
      if(this.items[i].getPath() == name)
        if(confirm("Overwrite saved playlist \"" + name + "\"?"))
          break;
        else
          return;
    if(name)
      backendController.save(name, bindFunction(this.saveDoneCallback, this));
}

SavedListsView.prototype.saveDoneCallback = function(data) {
    backendController.defaultCommandCallback(data);
    this.requestData();
}

inheritFrom(SavedListsView, ListView);

////////////////////////////////////////////////////////////////
function SavedPlaylistView(path, deleteCallback, domElement) {
    getSuper(arguments).apply(this, ["template_savedPlaylist", domElement]);

    this.path = path;
    this.deleteCallback = deleteCallback;

    this.getSubElement("addLink").onclick 
        = this.getSubElement("textContainer").onclick
        = bindFunction(backendController.load, backendController, [path, null]);
    this.getSubElement("deleteLink").onclick
        = bindFunction(this.requestDelete, this);

    setElementTextContent(this.getSubElement("textContainer"), basename(path));
}

SavedPlaylistView.prototype.requestDelete = function (data) {
    backendController.rm(this.path, bindFunction(this.deleteDoneCallback, this));
    addElementClass(this.domElement, "deleting"); 
    // This class is never removed, as there is no way to change our
    // path attribute, so we expect to be removed ourselves after the
    // playlist is gone.
}

SavedPlaylistView.prototype.deleteDoneCallback = function (data) {
    backendController.defaultCommandCallback(data);
    this.deleteCallback();
}

SavedPlaylistView.prototype.getPath = function (data) {
    return this.path;
}

inheritFrom(SavedPlaylistView, TemplateableElementWrapper);

////////////////////////////////////////////////////////////////
function PlaylistPositionListView(domElement) {
    getSuper(arguments).apply(this, arguments);
    this.domListElement = findElementsByClass(this.domElement, "listContainer").listContainer;
    this.separators = new Array();
    this.path = this.observerId = undefined;
}

PlaylistPositionListView.prototype.pushItem = function (item) {
    if(this.items.length > 0)
        this.separators.push(
            this.domListElement.appendChild(
                this.domElement.ownerDocument.createTextNode(", ")
            ));
    removeElementClass(this.domElement, "emptyPlaylistPositionList");
    getSuper(arguments).apply(this, arguments);
}
PlaylistPositionListView.prototype.popItem = function () {
    var separator = this.separators.pop();
    if(separator)
      separator.parentNode.removeChild(separator);
    var i = getSuper(arguments).apply(this, arguments);
    if(0 == this.items.length)
      addElementClass(this.domElement, "emptyPlaylistPositionList");
    return i;
}

PlaylistPositionListView.prototype.createItem = function () {
    return new PlaylistPosition("#");
}

PlaylistPositionListView.prototype.setPath = function (path) {
    if(undefined !== this.observerId)
        playlistPositionIndex.getFileObservable(this.path).removeObserver(this.observerId);
    
    this.path = path;
    var o = playlistPositionIndex.getFileObservable(this.path);
    this.observerId = o.addObserver(
        bindFunction(this.setValue, this)
        );
    this.setValue(o.getValue());
}

PlaylistPositionListView.prototype.setValue = function (positions) {
    var positionList = [];
    for(var p in positions)
        positionList.push(toNumber(p));
    positionList.sort(function (a,b) {return a-b;});
    this.setLength(positionList.length);
    for(var i=0; i<positionList.length; i++)
        this.items[i].setValue(positionList[i]+1);
}

inheritFrom(PlaylistPositionListView, ListView);

////////////////////////////////////////////////////////////////
function PlaylistPosition(textPrefix, domElement) {
    getSuper(arguments).apply(this, ["template_playlistPosition", domElement]);
    this.textPrefix = textPrefix;
}

PlaylistPosition.prototype.setValue = function (value) {
    setElementTextContent(this.domElement, this.textPrefix + value);
    setLocalLink(this.domElement, "#playlist;"+value);
}

inheritFrom(PlaylistPosition, TemplateableElementWrapper);

////////////////////////////////////////////////////////////////

function DirectoryContentView(domElement) {
    getSuper(arguments).apply(this, arguments);
    this.dataRequested = false;
    this.dataComplete = new Observable(false);
}

DirectoryContentView.prototype.requestData = function (path)
{
    addElementClass(this.domElement, "updating");
    this.setExpectedCallbackSequenceNumber(
        backendController.lsinfo(path, bindFunction(this.acceptCallbackData, this))
        );
    this.dataRequested = true;
}

DirectoryContentView.prototype.processCallbackDataTask = function()
{
    var status = getSuper(arguments).apply(this, arguments);

    switch(status)
    {
    case ContiguousList.PROCESSING_COMPLETE:
      this.dataComplete.setValue(true);
      break;
    case ContiguousList.PROCESSING_DEFERRED:
    case ContiguousList.PROCESSING_ABORTED:
      break;
    default:
      log.warn("Unexpected status code in DirectoryContentView.processCallbackDataTask");
      break;
    }
    return status;
}

DirectoryContentView.prototype.createItemFromCallbackData = function (entry)
{
    if(entry.directory)
      return new DirectoryView(entry.directory);
    return getSuper(arguments).apply(this, arguments);
}

DirectoryContentView.prototype.isCallbackDataItemInteresting = function(item)
{
    return item.directory || item.file ? true : false;
}


// one-shot callback mechanism to notify when directory content is loaded
DirectoryContentView.prototype.addDataCompleteSubscriber = function (callback)
{ 
    if(true == this.dataComplete.getValue())
      callback(true);
    else
      this.dataComplete.addOneShotObserver(callback);
}

DirectoryContentView.prototype.getItemByPath = function (path)
{
    for(var i=0; i<this.items.length; i++)
    {
	var entry = this.items[i];
	if(entry.getFilename && entry.getFilename() == path)
  	  return entry;
	else
	  if(entry.getPath && entry.getPath() == path) // TODO cleanup
	    return entry;
    }
    return null;
}

inheritFrom(DirectoryContentView, ListView);

////////////////////////////////////////////////////////////////
function ContiguousListModel() {
    ListModel.apply(this, arguments);
    getSuper(arguments).apply(this, arguments);
    this.length = new Observable();
}

ContiguousListModel.prototype.pushItem = function (item) {
    getSuper(arguments).apply(this, arguments);
    this.length.setValue(this.items.length);
}

ContiguousListModel.prototype.popItem = function (item) {
    var i = getSuper(arguments).apply(this, arguments);
    this.length.setValue(this.items.length);
    return i;
}

ContiguousListModel.prototype.setLength = function (l, fastShorten) {
    getSuper(arguments).apply(this, arguments);
    this.length.setValue(this.items.length); //needed in case of fastShorten
}

ContiguousListModel.prototype.getItemModel = function (i) {
    return this.items[i];
}

ContiguousListModel.prototype.createItem = function () {
    return new ObservableModel();
}

ContiguousListModel.prototype.createItemFromCallbackData = function(entry) {
    var d = this.createItem();
    for(var a in entry)
        d.setAttribute(a, entry[a]);
    return d;
}

inheritFrom(ContiguousListModel, ListModel);
inheritFrom(ContiguousListModel, ContiguousList);

////////////////////////////////////////////////////////////////
function ServerSearch() {
    getSuper(arguments).apply(this, arguments);
    this.backend = null;
    this.queryText = this.queryType = null;
    this.dataComplete.setValue(true);
}

ServerSearch.prototype.useBackend = function (backend) { // TODO duplication
    if(this.backend)
      log.warn("calling useBackend twice is unlikely to work, as it doesn't invalidate old data");
    this.backend = backend;
}

ServerSearch.prototype.setQuery = function (query, type) {
    if(query != this.queryText || type != this.queryType)
    {
        this.setLength(0,true);
        this.queryText = query;
        this.queryType = type;
        if(this.queryText && this.queryType)
            this.submitQuery();
    }
}

ServerSearch.prototype.getQueryText = function () {
    return this.queryText;
}

ServerSearch.prototype.getQueryType = function () {
    return this.queryType;
}

ServerSearch.prototype.submitQuery = function () {
    this.dataComplete.setValue(false);
    
    log.info('Sending query "' + this.queryType 
             + '","' + this.queryText + '" to server');
    
    this.setExpectedCallbackSequenceNumber(
        this.backend.search([this.queryType, this.queryText],
                            bindFunction(this.acceptCallbackData, this)
            ));
}

ServerSearch.prototype.processCallbackDataTask = function() {
    var status = getSuper(arguments).apply(this, arguments);
    if(ContiguousList.PROCESSING_COMPLETE === status)
        this.dataComplete.setValue(true);

    return status;
}

inheritFrom(ServerSearch, ContiguousListModel);

////////////////////////////////////////////////////////////////
function SearchModel() {
    getSuper(arguments).apply(this, arguments);
    this.queryText = this.queryType = null;
    this.queryTextWords = new Array();
    this.queryTextWordREs = new Array();
    this.dataComplete.setValue(true);

    this.serverSearch = new ServerSearch();
    // watch for new matches
    this.serverSearch.length.addObserver(bindFunction(this.startResultFetch, this));
    // watch for the end of the matches
    this.serverSearch.dataComplete.addObserver(bindFunction(this.startResultFetch, this));

    this.serverItemsSeen = 0;
    this.resultFetchTaskId = null;
}

SearchModel.prototype.useBackend = function (backend) {
    this.serverSearch.useBackend(backend);
}

SearchModel.prototype.setQuery = function (query, type) {
    this.stopResultFetch();
    this.dataComplete.setValue(false);
    this.serverItemsSeen = 0;
    this.setLength(0, true);

    this.queryText = query ? query : "";
    this.queryType = type;

    this.queryTextWords = this.queryText.split(/\s+/);
    this.queryTextWordREs = [];
    for(var i=0; i<this.queryTextWords.length; i++)
      this.queryTextWordREs.push(new RegExp(this.queryTextWords[i], "i"));

    if(this.isAutoSubmittable() && !this.isPruneable())
      this.submitQuery();
    else
      this.startResultFetch();
}

SearchModel.autoSubmitThreshold = 3;

SearchModel.prototype.isAutoSubmittable = function () {
    for(var i=0; i<this.queryTextWords.length; i++)
      if(this.queryTextWords[i].length >= SearchModel.autoSubmitThreshold)
        return true;
    return false;
}

SearchModel.prototype.isPruneable = function () {
    // are the matches to this query a subset of the last server query?
    return this.queryText
           && this.queryType === this.serverSearch.getQueryType()
           && 0 <= this.queryText.indexOf(this.serverSearch.getQueryText());
}

SearchModel.prototype.submitQuery = function () {
    // The server can only do literal string searches, but we want to
    // match each word individually, so pick the longest word and send
    // that to the server search.
    var longestWord = "";
    for(var i=0; i<this.queryTextWords.length; i++)
      if(this.queryTextWords[i].length > longestWord.length)
        longestWord = this.queryTextWords[i];

    this.serverSearch.setQuery(longestWord, this.queryType);
}

SearchModel.prototype.startResultFetch = function () {
    if(null === this.resultFetchTaskId)
    {
        this.resultFetchTaskId = scheduler.createTask(
            bindFunction(this.fetchResultTask, this)
            );
    }
}
SearchModel.prototype.stopResultFetch = function () {
    if(null !== this.resultFetchTaskId)
        scheduler.killTask(this.resultFetchTaskId);
    this.resultFetchTaskId = null;
}

SearchModel.prototype.fetchResultTask = function () {
    var serverMatchCount = this.serverSearch.length.getValue();
    var fetchedCount = this.getLength();

    for(; this.serverItemsSeen<serverMatchCount; this.serverItemsSeen++)
    {
        var serverMatch = this.serverSearch.getItem(this.serverItemsSeen);
        if(this.itemMatches(serverMatch))
        {
            this.setLength(fetchedCount + 1);
            this.getItem(fetchedCount).copyModel(serverMatch);
            fetchedCount++;
        }
	if(scheduler.isTimeSliceExpired())
	{
            scheduler.requestRestart();
	    return;
	}
    }

    if(this.serverSearch.dataComplete.getValue())
        this.dataComplete.setValue(true);

    this.resultFetchTaskId = null;
}


SearchModel.prototype.itemMatches = function (item) {
    if(!this.isPruneable())
      return false;

    // inconsistent naming in the mpd protocol
    var attributeNames = {
	"album": "Album",
	"artist": "Artist",
	"filename" : "file", 
	"title": "Title"
    };
    var attribute = item.getAttribute(attributeNames[this.queryType]);

    for(var r=0; r<this.queryTextWordREs.length; r++)
      if(!this.queryTextWordREs[r].test(attribute))
    {
        return false;
    }
    
    return true;
}

inheritFrom(SearchModel, ContiguousListModel);

////////////////////////////////////////////////////////////////

function SearchView(domElement) {
    getSuper(arguments).apply(this, arguments); // Warning: this will
                                                // call ElementWrapper()
                                                // again.
    this.domListElement = this.getSubElement("searchResults");

    this.getSubElement("searchInput").onchange
        = bindFunction(this.queryTextChanged, this);
    // hack to ensure onchange handler gets called
    // TODO onchange doesn't seem to fire in konqueror
    this.getSubElement("searchInput").onkeyup 
        = new Function("this.blur(); this.focus();")

    this.getSubElement("typeSelect").onchange
        = bindFunction(this.typeChanged, this);

    this.getSubElement("addLink").onclick
        = bindFunction(this.addFilesToPlaylist, this);

    this.queryText = this.serverQuery = null;
    this.queryType = this.getSubElement("typeSelect").value;
    this.queryTextWords = [];
    this.queryTextWordREs = [];

    // TODO submit onenter even if threshold not met
    this.search = new SearchModel();
    this.search.useBackend(backendController);
    this.observeModel(this.search);
}

SearchView.prototype.acceptModelLength = function(l) {
    getSuper(arguments).apply(this, arguments);
    setElementTextContent(this.getSubElement("matchCount"), l);
}

SearchView.prototype.createItem = function () {
    return new FileView();
}

SearchView.prototype.queryTextChanged = function () {
    this.queryText = this.getSubElement("searchInput").value;
    this.search.setQuery(this.queryText, this.queryType);
}

SearchView.prototype.typeChanged = function () {
    this.queryType = this.getSubElement("typeSelect").value;
    this.search.setQuery(this.queryText, this.queryType);
}

SearchView.prototype.addFilesToPlaylist =  function () {
    var filenames = new Array();
    this.search.forEachItem(function(i) {
	filenames.push(i.getAttribute("file"));
    });
    backendController.add(filenames);
}

inheritFrom(SearchView, ScrollableListModelView);

////////////////////////////////////////////////////////////////
function BackendMonitor(name) {
    this.name = name;
    this.monitorFrame = document.body.appendChild(
	document.createElement("iframe")
	);
    setElementDisplay(this.monitorFrame, false);
    // keep this value larger than the HEARTBEAT_INTERVAL in mpdmonitor.cgi
    this.watchdogInterval = 10*1000; 
    this.observable = new Observable();
}

BackendMonitor.prototype.start = function() {
    this.monitorFrame.src = "mpdmonitor.cgi?callback=parent."+this.name+".monitorCallback";
    this.resetWatchdog();
}

BackendMonitor.prototype.stop = function() {
    if(this.watchdogTimerId)
      window.clearTimeout(this.watchdogTimerId);
    this.monitorFrame.src = "about:blank";
}

BackendMonitor.prototype.getObservable = function() {
    return this.observable;
}

BackendMonitor.prototype.monitorCallback = function(key, value) {
    this.resetWatchdog();
    this.observable.setValue([key, value])
}

BackendMonitor.prototype.resetWatchdog = function() {
    if(this.watchdogTimerId)
      window.clearTimeout(this.watchdogTimerId);
    this.watchdogTimerId = window.setTimeout(
	bindFunction(this.watchdogCallback, this), this.watchdogInterval
	);
}

BackendMonitor.prototype.watchdogCallback = function() {
    log.warn("No activity from backend monitor process, restarting it.");
    if(this.monitorFrame.contentWindow)
      this.monitorFrame.contentWindow.location.reload();
    else
      log.error("Can't restart backend monitor process. Try refreshing the page.");
    this.watchdogInterval *= 1.5;
    log.info("Increased watchdog interval to " + this.watchdogInterval / 1000 + "s");
    this.resetWatchdog();
}


////////////////////////////////////////////////////////////////
// Client-originating interaction with MPD

function BackendController() {
    this.nextCommandSequenceNumber = 0;
}

BackendController.prototype = {
    // These will be pointed to functions by the code below
    play: "",
    pause: "",
    stop: "",
    next: "",
    prev: "",
    set_repeat: "",
    set_fade: "",
    set_random: "",
    shuffle: "",
    clear: "",
    playid: "",
    deleteid: "",
    set_volume: "",
    seek: "",
    lsinfo: "",
    add: "",
    search: "",
    load: "",
    save: "",
    rm: "",
    playlist_changes: ""
}

for(var name in BackendController.prototype)
{
    BackendController.prototype[name] = new Function ("", "return this.sendMpdCommand('"+name+"', arguments[0], arguments[1])");
}

BackendController.prototype.defaultCommandCallback = function(data) {
   if(data.error)
      log.error("MPD backend error: "+ data.error.command + ":" + data.error.message);
}

BackendController.prototype.sendMpdCommand = function (command, args, callback) {
    function cleanValue(a)
    {
      if("number" === typeof a)
          return Math.round(a);
      else
          return a;
    }

    var c = { "command": command };

    if(undefined !== args && null !== args)
    {
        if(args.length && "object" === typeof args)
        {
            for(var i=0; i<args.length; i++)
                c[i] = cleanValue(args[i]);
        }
        else
            c[0] = cleanValue(args);
    }

    if("function" !== typeof callback)
        callback = this.defaultCommandCallback;

    return this.sendMpdQuery(c, callback);
}

BackendController.prototype.sendMpdQuery = function (parameters, callback)
{
    var url='mpdcommand.cgi';
    var sequenceNumber = this.nextCommandSequenceNumber;
    this.nextCommandSequenceNumber++;
    var queryString = "";
    for(var p in parameters)
	queryString += escape(p) + "=" + escape(parameters[p]) + "&";

    var xmlhttp = new XMLHttpRequest();
    xmlhttp.open("POST", url, true);
    xmlhttp.onreadystatechange = function () {
	log.info("XMLHttpRequest number " + sequenceNumber
		 + " readyState == " + xmlhttp.readyState);
	if(4 == xmlhttp.readyState)
	{
            var data;
            try {
                data = eval("(" + xmlhttp.responseText + ")");
            } catch (e) {
                data = { error : { message : "Could not parse the response from the web server when retrieving '"+url+"' with query string '"+queryString+"'. Check your server logs for errors related to this URL. The JavaScript error reported was '"+e+"'. Response text was: " + xmlhttp.responseText }};
            }
            data.sequenceNumber = sequenceNumber;
            callback(data);
            // TODO catch http errors 
	}
    };
    xmlhttp.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");
    log.info("Sending '"+queryString+"' asXMLHttpRequest number " + sequenceNumber);
    xmlhttp.send(queryString);
    return sequenceNumber;
}

////////////////////////////////////////////////////////////////

function TabSet()
{
    this.elements = new Object();
    for(var i=0; i<arguments.length; i++)
    {
	var name = arguments[i];
	this.elements[name] = new Object();
	this.elements[name].tab = getElementById("tab_" + name);
	this.elements[name].pane = getElementById("pane_" + name);
	this.elements[name].tab.onclick = 
	    bindFunction(this.showElement, this, [name]);
    }
    this.shownElement = "";
    this.showElement(arguments[0]);
}

TabSet.prototype.showElement = function (name)
{
    if(this.shownElement != name)
    {
        if(this.shownElement)
          this.elements[this.shownElement].scrollPosition = window.scrollY;
	for(var n in this.elements)
	{
	    setToggleElementClass(this.elements[n].tab, (name == n));
	    this.elements[n].tab.blur(); // TODO hacky
            setElementDisplay(this.elements[n].pane, (name == n));
	}
	this.shownElement = name;
        
        if(navigator.appName.match("Netscape"))
          forcePageReflow();  // Without this, the vertical height of the page is not
                              // recalculated when switching from a taller to a smaller
                              // tab pane (seen with Gecko/20050927)

        var savedPosition = this.elements[this.shownElement].scrollPosition;
        if(savedPosition !== undefined)
          window.scrollTo(window.scrollX, savedPosition);
    }

    // Don't allow konqueror to follow the href on the link that
    // triggered us (would cause the iframe to stop loading) 
    // TODO work around this
    return !(navigator.appName.match("Konqueror"));
}

////////////////////////////////////////////////////////////////

// TODO do something intelligent with the location when the user
// scrolls the page. Probably best to discard all but the tab id and
// let TabSet restore the scroll position

function LocationWatcher() {
    this.lastLocationHash = "";
    this.locationPollTimer = window.setInterval(
	bindFunction(this.pollLocation, this),
	LocationWatcher.locationPollInterval
	);
}

LocationWatcher.locationPollInterval = 500;

LocationWatcher.prototype.pollLocation = function () {
    if(navigator.appName.match("Konqueror"))
      return "TODO find a workaround";

    var locationHash = window.location.hash;
    if(locationHash == this.lastLocationHash) // only react to changing hash values
      return undefined;
    this.setLocation(locationHash);
    return undefined;
}

LocationWatcher.prototype.setLocation = function (locationHash) {
    this.lastLocationHash = locationHash;
    if(!navigator.appName.match("Konqueror"))
      window.location.hash = locationHash.replace("#", "");

    var matches = locationHash.match(/#?([^;]*)(;(.*))?$/);
    var tab = matches[1];

    tabs.showElement(tab);

    if(matches[3])
    {
	var item = unescape(matches[3]);
        switch(tab)
        {
            case "browse":
                databaseView.showFile(item);
                break;
            case "playlist":
                playlistView.showItem(item);
                break;
        }
    }
    
    return true; // return value needed for onclick handlers
}

////////////////////////////////////////////////////////////////

function StickableElement(domElement)
{
    getSuper(arguments).apply(this, arguments);
    this.getSubElement("stickLink").onclick = bindFunction(this.toggleStickiness, this);
    this.bodyOffset = 0;

    domElement.style.left 
        = domElement.style.right 
        = getElementXPosition(domElement) + "px"; // assumption: element is centred

    // stretch element to top of page, so no scrolled content shows through above it
    domElement.style.paddingTop = getElementYPosition(domElement) + "px";
    document.body.style.marginTop = "0px";
    domElement.style.top = "0px";

    this.stuck = false;
    if(true.toString() == readCookie(StickableElement.cookieName))
        this.toggleStickiness();
}

StickableElement.offsetAdjustInterval = 1000;
StickableElement.cookieName = "webmpd-stickyControls";

StickableElement.prototype.toggleStickiness = function() {
    this.stuck = !this.stuck;
    if(this.stuck)
    {
        this.domElement.style.position = "fixed";
        this.adjustBodyOffset();
        // The offset will need to be readjusted for user font-size changes
        // or text reflowing within the stickable element.
        this.offsetTimerId = window.setInterval(
            bindFunction(this.adjustBodyOffset, this),
            StickableElement.offsetAdjustInterval);
    }
    else
    {
        window.clearInterval(this.offsetTimerId);
        this.setBodyOffset(0);
        this.domElement.style.position = "";
        scrollToElement(this.domElement);
    }

    setToggleElementClass(this.getSubElement("stickLink"), this.stuck);

    createCookie(StickableElement.cookieName, this.stuck, 7);
}

StickableElement.prototype.adjustBodyOffset = function() {
    this.setBodyOffset(this.domElement.offsetHeight);
}

StickableElement.prototype.setBodyOffset = function(offset) {
    this.bodyOffset = offset;
    document.body.style.marginTop = offset + "px";
}

StickableElement.prototype.getBodyOffset = function() {
    return this.bodyOffset;
}

inheritFrom(StickableElement, ElementWrapper);


////////////////////////////////////////////////////////////////
var log;
var backendController;
var playerView;
var playlistModel;
var playlistView;
var databaseView;
var tabs;
var locationWatcher;
var stickable;
var playlistPositionIndex;
var backendMonitor;
var scheduler;

function init() {
  log = new Log(getElementById("messages"));
  scheduler = new Scheduler();
  scheduler.start();

  backendController = new BackendController();
  locationWatcher = new LocationWatcher();

  backendMonitor = new BackendMonitor("backendMonitor");
  var playerStatusModel = new PlayerStatusModel();
  playerStatusModel.observeBackendMonitor(backendMonitor);
  var playerView = new PlayerView(getElementById("player"));
  playerView.observeModel(playerStatusModel);
  
  playlistModel = new PlaylistModel();
  playlistModel.useBackend(backendController);
  playlistModel.observePlayerStatusModel(playerStatusModel);

  playlistPositionIndex = new PlaylistPositionIndex();
  playlistPositionIndex.observePlaylistModel(playlistModel);
  
  playlistView = new PlaylistView(getElementById("pane_playlist"));
  playlistView.observeModel(playlistModel);

  backendMonitor.start();

  databaseView = new DirectoryView("", getElementById("pane_browse"));
  var savedListsView = new SavedListsView(getElementById("pane_lists"));
  tabs = new TabSet("playlist", "browse", "search", "lists");
  new SearchView(getElementById("pane_search"));
  stickable = new StickableElement(getElementById("stickableSection"));
  new StyleSwitcher(getElementById("styleSelect"));
  databaseView.expand();
  savedListsView.requestData();
}  

